#include <config.h>
#include <array>
#include <vector>
#include <fstream>

#include <iostream>
#include <dune/common/indices.hh>
#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/float_cmp.hh>
#include <dune/common/math.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/spqr.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/io.hh>

#include <dune/functions/functionspacebases/interpolate.hh>

#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>

#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/periodicbasis.hh>

#include <dune/functions/functionspacebases/subspacebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>


#include <dune/microstructure/prestrainimp.hh>
#include <dune/functions/functionspacebases/hierarchicvectorwrapper.hh>

// #include <dune/fufem/discretizationerror.hh>

using namespace Dune;



template<class LocalView, class Matrix, class localFunction1, class localFunction2>
void computeElementStiffnessMatrix(const LocalView& localView,
                                   Matrix& elementMatrix,
                                   const localFunction1& mu,
                                   const localFunction2& lambda,
                                   const double gamma
                                   )
{

  // Local StiffnessMatrix of the form:
  // | phi*phi    m*phi |
  // | phi *m     m*m   |

  using Element = typename LocalView::Element;
  const Element element = localView.element();
  auto geometry = element.geometry();

  constexpr int dim = Element::dimension;
  constexpr int nCompo = dim;

  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;
  //     using Domain = typename LocalView::GridView::Codim<0>::Geometry::GlobalCoordinate;
  //     using FuncScalar = std::function< ScalarRT(const Domain&) >;
  //     using Func2Tensor = std::function< Matload_alpha1 ,rixRT(const Domain&) >;


  elementMatrix.setSize(localView.size()+3, localView.size()+3);         //extend by dim ´R_sym^{2x2}
  elementMatrix = 0;


  // LocalBasis-Offset
  const int localPhiOffset = localView.size();

  const auto& localFiniteElement = localView.tree().child(0).finiteElement();     // Unterscheidung children notwendig?
  const auto nSf = localFiniteElement.localBasis().size();

//   std::cout << "localView.size(): " << localView.size() << std::endl; 
//   std::cout << "nSf : " << nSf << std::endl; 

  ///////////////////////////////////////////////
  // Basis for R_sym^{2x2}            // wird nicht als Funktion benötigt da konstant...
  //////////////////////////////////////////////
  MatrixRT G_1 {{1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0, 0.0}};
  MatrixRT G_2 {{0.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0, 0.0, 0.0}};
  MatrixRT G_3 {{0.0, 1.0/sqrt(2), 0.0}, {1.0/sqrt(2), 0.0, 0.0}, {0.0, 0.0, 0.0}};
  std::array<MatrixRT,3 > basisContainer = {G_1, G_2, G_3};
  //print:
  //     printmatrix(std::cout, basisContainer[0] , "G_1", "--");
  //     printmatrix(std::cout, basisContainer[1] , "G_2", "--");
//       printmatrix(std::cout, basisContainer[2] , "G_3", "--");
  ////////////////////////////////////////////////////

//   int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1)+5;  // TEST 
  int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);
  const auto& quad = QuadratureRules<double,dim>::rule(element.type(), orderQR);


  for (const auto& quadPoint : quad)
  {

    const auto& quadPos = quadPoint.position();
    const auto jacobianInverseTransposed = geometry.jacobianInverseTransposed(quadPoint.position());
    const auto integrationElement = geometry.integrationElement(quadPoint.position());

    std::vector< FieldMatrix< double, 1, dim> > referenceGradients;
    localFiniteElement.localBasis().evaluateJacobian(quadPoint.position(),
                                                     referenceGradients);

    // Compute the shape function gradients on the grid element
    std::vector<FieldVector<double,dim> > gradients(referenceGradients.size());
    //         std::vector< VectorRT> gradients(referenceGradients.size());

    for (size_t i=0; i<gradients.size(); i++)
      jacobianInverseTransposed.mv(referenceGradients[i][0], gradients[i]);


    // "phi*phi"-part
    for (size_t k=0; k < nCompo; k++)
      for (size_t l=0; l< nCompo; l++)
      {
        for (size_t i=0; i < nSf; i++)
          for (size_t j=0; j < nSf; j++ )
          {

            // (scaled) Deformation gradient of the ansatz basis function
            MatrixRT defGradientU(0);
            defGradientU[k][0] = gradients[i][0];                       // Y
            defGradientU[k][1] = gradients[i][1];                       //X2
            defGradientU[k][2] = (1.0/gamma)*gradients[i][2];           //X3
//             printmatrix(std::cout, defGradientU , "defGradientU", "--");
            // (scaled)  Deformation gradient of the test basis function
            MatrixRT defGradientV(0);
            defGradientV[l][0] = gradients[j][0];                       // Y
            defGradientV[l][1] = gradients[j][1];                       //X2
            defGradientV[l][2] = (1.0/gamma)*gradients[j][2];           //X3


            // symmetric Gradient (Elastic Strains)
            MatrixRT strainU, strainV;
            for (int ii=0; ii<nCompo; ii++)
              for (int jj=0; jj<nCompo; jj++)
              {
                strainU[ii][jj] = 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);                 // symmetric gradient
                strainV[ii][jj] = 0.5 * (defGradientV[ii][jj] + defGradientV[jj][ii]);
                //                     strainV[ii][jj] = 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);         // same ? genügt strainU
//                 printmatrix(std::cout, strainU , "strainU", "--");
              }

            // St.Venant-Kirchhoff stress
            // < L sym[D_gamma*nabla phi_i], sym[D_gamma*nabla phi_j] >
            // stressU*strainV
            MatrixRT stressU(0);
            stressU.axpy(2*mu(quadPos), strainU);         //this += 2mu *strainU

            double trace = 0;
            for (int ii=0; ii<nCompo; ii++)
              trace += strainU[ii][ii];

            for (int ii=0; ii<nCompo; ii++)
              stressU[ii][ii] += lambda(quadPos) * trace;

            // Local energy density: stress times strain
            double energyDensity = 0;
            for (int ii=0; ii<nCompo; ii++)
              for (int jj=0; jj<nCompo; jj++)
                energyDensity += stressU[ii][jj] * strainV[ii][jj];                      // "phi*phi"-part
            //                     energyDensity += stressU[ii][jj] * strainU[ii][jj];

            //                     size_t row = localView.tree().child(k).localIndex(i);             // kann auf Unterscheidung zwischen k & l verzichten?
            //                     size_t col = localView.tree().child(l).localIndex(j);             // siehe DUNE-book p.394
            //                 size_t col = localView.tree().child(k).localIndex(j);             // Indizes mit k=l genügen .. Kroenecker-Delta_kl  NEIN???
            size_t col = localView.tree().child(k).localIndex(i);                        // kann auf Unterscheidung zwischen k & l verzichten?!
            size_t row = localView.tree().child(l).localIndex(j);

            elementMatrix[row][col] += energyDensity * quadPoint.weight() * integrationElement;

          }

      }


    // "m*phi"   & "phi*m" -part
    for (size_t l=0; l< nCompo; l++)
      for (size_t j=0; j < nSf; j++ )
      {


        // (scaled)  Deformation gradient of the test basis function
        MatrixRT defGradientV(0);
        defGradientV[l][0] = gradients[j][0];                       // Y
        defGradientV[l][1] = gradients[j][1];                       //X2
        defGradientV[l][2] = (1.0/gamma)*gradients[j][2];           //X3

        // symmetric Gradient (Elastic Strains)
        MatrixRT strainV;
        for (int ii=0; ii<nCompo; ii++)
          for (int jj=0; jj<nCompo; jj++)
          {
            strainV[ii][jj] = 0.5 * (defGradientV[ii][jj] + defGradientV[jj][ii]);
          }


        for( size_t m=0; m<3; m++)
        {

          // < L G_i, sym[D_gamma*nabla phi_j] >
          // L G_i* strainV

          // St.Venant-Kirchhoff stress
          MatrixRT stressG(0);
          stressG.axpy(2*mu(quadPos), basisContainer[m]);           //this += 2mu *strainU

          double traceG = 0.0;
          for (int ii=0; ii<nCompo; ii++)
          {
//             std::cout << basisContainer[m][ii][ii] << std::endl;
            traceG += basisContainer[m][ii][ii];
          }

          for (int ii=0; ii<nCompo; ii++)
            stressG[ii][ii] += lambda(quadPos) * traceG;

          double energyDensityGphi = 0.0;
          for (int ii=0; ii<nCompo; ii++)
            for (int jj=0; jj<nCompo; jj++)
              energyDensityGphi += stressG[ii][jj] * strainV[ii][jj];                       // "m*phi"-part

          size_t row = localView.tree().child(l).localIndex(j);

          auto value = energyDensityGphi * quadPoint.weight() * integrationElement;

          elementMatrix[row][localPhiOffset+m] += value;
          elementMatrix[localPhiOffset+m][row] += value;

        }

      }



    // "phi*m"-part
//     for (size_t k=0; k < nCompo; k++)
//         for (size_t i=0; i < nSf; i++)
//         {
// 
//             // (scaled) Deformation gradient of the ansatz basis function
//             MatrixRT defGradientU(0);
//             defGradientU[k][0] = gradients[i][0];               // Y
//             defGradientU[k][1] = gradients[i][1];               //X2
//             defGradientU[k][2] = (1.0/gamma)*gradients[i][2];   //X3
// 
// 
//             // symmetric Gradient (Elastic Strains)
//             MatrixRT strainU;
//             for (int ii=0; ii<nCompo; ii++)
//             for (int jj=0; jj<nCompo; jj++)
//             {
//                 strainU[ii][jj] = 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);         // symmetric gradient
//             }
// 
//             // St.Venant-Kirchhoff stress
//             MatrixRT stressU(0);
//             stressU.axpy(2*mu(quadPos), strainU); //this += 2mu *strainU
// 
//             double trace = 0;
//             for (int ii=0; ii<nCompo; ii++)
//             trace += strainU[ii][ii];
// 
//             for (int ii=0; ii<nCompo; ii++)
//             stressU[ii][ii] += lambda(quadPos) * trace;
// 
//             for( size_t n=0; n<3; n++)
//             {
// 
//                 // < L sym[D_gamma*nabla phi_i], G_j >
//                 double energyDensityPhiG = 0;
//                 for (int ii=0; ii<nCompo; ii++)
//                 for (int jj=0; jj<nCompo; jj++)
//                     energyDensityPhiG += stressU[ii][jj] * basisContainer[n][ii][jj];             // "phi*m"-part
// 
//                 size_t col = localView.tree().child(k).localIndex(i);
// 
//                 elementMatrix[localPhiOffset+n][col] += energyDensityPhiG * quadPoint.weight() * integrationElement;
// 
//             }
//         }
// 



    // "m*m"-part
    for(size_t m=0; m<3; m++)
      for(size_t n=0; n<3; n++)
      {

        // St.Venant-Kirchhoff stress
        MatrixRT stressG(0);
        stressG.axpy(2*mu(quadPos), basisContainer[m]);         //this += 2mu *strainU

        double traceG = 0.0;
        for (int ii=0; ii<nCompo; ii++)
          traceG += basisContainer[m][ii][ii];

        for (int ii=0; ii<nCompo; ii++)
          stressG[ii][ii] += lambda(quadPos) * traceG;

        double energyDensityGG = 0.0;

        for (int ii=0; ii<nCompo; ii++)
          for (int jj=0; jj<nCompo; jj++)
            energyDensityGG += stressG[ii][jj] * basisContainer[n][ii][jj];                     // "m*m"-part

        elementMatrix[localPhiOffset+m][localPhiOffset+n]= energyDensityGG * quadPoint.weight() * integrationElement;

      }

  }


}




// Get the occupation pattern of the stiffness matrix
template<class Basis>
void getOccupationPattern(const Basis& basis, MatrixIndexSet& nb)
{
  //  OccupationPattern:
  // | phi*phi    m*phi |
  // | phi *m     m*m   |
  auto gridView = basis.gridView();

  auto localView = basis.localView();

  const int phiOffset = basis.dimension();

  nb.resize(basis.size()+3, basis.size()+3);       //

  for (const auto& element : elements(gridView))
  {
    localView.bind(element);
    for (size_t i=0; i<localView.size(); i++)
    {
      // The global index of the i-th vertex of the element
      auto row = localView.index(i);
      for (size_t j=0; j<localView.size(); j++ )
      {
        // The global index of the j-th vertex of the element
        auto col = localView.index(j);

        nb.add(row[0],col[0]);                   // nun würde auch nb.add(row,col) gehen..
      }
    }


    for (size_t i=0; i<localView.size(); i++)
    {
      auto row = localView.index(i);

      for (size_t j=0; j<3; j++ )
      {
        nb.add(row,phiOffset+j);               // m*phi part of StiffnessMatrix

        nb.add(phiOffset+j,row);               // phi*m part of StiffnessMatrix
      }
    }

    for (size_t i=0; i<3; i++ )
      for (size_t j=0; j<3; j++ )
      {
        nb.add(phiOffset+i,phiOffset+j);               // m*m part of StiffnessMatrix
      }

  }

  //////////////////////////////////////////////////////////////////
  // setIntegralZero:           
  //////////////////////////////////////////////////////////////////
  int arbitraryIndex = 0;

  FieldVector<int,3> row;
  unsigned int elementCounter = 1;

//   for (const auto& element : elements(basis.gridView()))
//   {
//     localView.bind(element);
// 
//     if(elementCounter == 1)             // get arbitraryIndex(global) for each Component        ..alternativ:gridView.indexSet
//     {
//       for (int k = 0; k < 3; k++)
//       {
//         auto rowLocal = localView.tree().child(k).localIndex(arbitraryIndex);
//         row[k] = localView.index(rowLocal);
//       }
//     }
// 
//     // "global assembly"
//     for (int k = 0; k<3; k++)
//       for (size_t j=0; j<localView.size(); j++)
//       {
//         auto col = localView.index(j);
//         nb.add(row[k],col);
//       }
//     elementCounter++;
//   }
  /////////////////////////////////////////////////////////////////

  ////////////////////////////////////////////////////////////////
  // setOneBaseFunctionToZero         necessary?                //TODO  besser: if(parameterSet_.get<bool>("set_one_base_function_to_0" )){ 

  //Determine 3 global indices (one for each component of an arbitrary FE-function)..    oder mittels Offset?
  // use first element:
  for (const auto& element : elements(basis.gridView()))
  {
    localView.bind(element);

    if(elementCounter == 1)             // get arbitraryIndex(global) for each Component        ..alternativ:gridView.indexSet
    {
      for (int k = 0; k < 3; k++)
      {
        auto rowLocal = localView.tree().child(k).localIndex(arbitraryIndex);
        row[k] = localView.index(rowLocal);
      }
    }

    // "global assembly"
    for (int k = 0; k<3; k++)
        nb.add(row[k],row[k]);
      
    elementCounter++;
  }


}




// Compute the source term for a single element
// < L (sym[D_gamma*nabla phi_i] + M_i ), x_3G_alpha >
template<class LocalView, class LocalFunction1, class LocalFunction2, class Vector, class Force>
void computeElementLoadVector( const LocalView& localView,
                               LocalFunction1& mu,
                               LocalFunction2& lambda,
                               const double gamma,
                               Vector& elementRhs,
                               const Force& forceTerm
                               )
{
  // | f*phi|
  // | ---  |
  // | f*m  |

  using Element = typename LocalView::Element;
  const auto element = localView.element();
  const auto geometry = element.geometry();
  constexpr int dim = Element::dimension;
  constexpr int nCompo = dim;

  using VectorRT = FieldVector< double, nCompo>;
  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;

  // Set of shape functions for a single element
  const auto& localFiniteElement= localView.tree().child(0).finiteElement();
  const auto nSf = localFiniteElement.localBasis().size();
  //     const auto& localFiniteElement = localView.tree().finiteElement();


  elementRhs.resize(localView.size() +3);
  elementRhs = 0;


  // LocalBasis-Offset
  const int localPhiOffset = localView.size();

  ///////////////////////////////////////////////
  // Basis for R_sym^{2x2}
  //////////////////////////////////////////////
  MatrixRT G_1 {{1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0, 0.0}};
  MatrixRT G_2 {{0.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0, 0.0, 0.0}};
  MatrixRT G_3 {{0.0, 1.0/sqrt(2), 0.0}, {1.0/sqrt(2), 0.0, 0.0}, {0.0, 0.0, 0.0}};
  std::array<MatrixRT,3 > basisContainer = {G_1, G_2, G_3};

//   int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1)+5;  // TEST 
  int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);   // für simplex
  const auto& quad = QuadratureRules<double,dim>::rule(element.type(), orderQR);


  for (const auto& quadPoint : quad) 
  {


    const FieldVector<double,dim>& quadPos = quadPoint.position();
    const auto jacobian = geometry.jacobianInverseTransposed(quadPos);
    const double integrationElement = geometry.integrationElement(quadPos);

    std::vector<FieldMatrix<double,1,nCompo> > referenceGradients;
    localFiniteElement.localBasis().evaluateJacobian(quadPos, referenceGradients);

    std::vector< VectorRT > gradients(referenceGradients.size());
    for (size_t i=0; i< gradients.size(); i++)
      jacobian.mv(referenceGradients[i][0], gradients[i]);


    // "f*phi"-part
    for (size_t i=0; i < nSf; i++)
      for (size_t k=0; k < nCompo; k++)
      {
        // Deformation gradient of the ansatz basis function
        MatrixRT defGradientV(0);
        defGradientV[k][0] = gradients[i][0];                                 // Y
        defGradientV[k][1] = gradients[i][1];                                 // X2
        defGradientV[k][2] = (1.0/gamma)*gradients[i][2];                     // X3

        // Elastic Strain
        MatrixRT strainV;                                                       //strainV never used?
        for (int ii=0; ii<nCompo; ii++)
          for (int jj=0; jj<nCompo; jj++)
            strainV[ii][jj] = 0.5 * (defGradientV[ii][jj] + defGradientV[jj][ii]);

        // St.Venant-Kirchhoff stress
        MatrixRT stressV(0);
        stressV.axpy(2*mu(quadPos), strainV);                 //this += 2mu *strainU

        double trace = 0.0;
        for (int ii=0; ii<nCompo; ii++)
          trace += strainV[ii][ii];

        for (int ii=0; ii<nCompo; ii++)
          stressV[ii][ii] += lambda(quadPos) * trace;


        // Local energy density: stress times strain
        double energyDensity = 0.0;
        for (int ii=0; ii<nCompo; ii++)
          for (int jj=0; jj<nCompo; jj++)
            energyDensity += stressV[ii][jj] *forceTerm(quadPos)[ii][jj];                                              


        size_t row = localView.tree().child(k).localIndex(i);
        elementRhs[row] += energyDensity * quadPoint.weight() * integrationElement;
      }


    // "f*m"-part
    for (size_t m=0; m<3; m++)
    {

      // St.Venant-Kirchhoff stress
      MatrixRT stressG(0);
      stressG.axpy(2*mu(quadPos), basisContainer[m]);       //this += 2mu *strainU

      double traceG = 0;
      for (int ii=0; ii<nCompo; ii++)
        traceG += basisContainer[m][ii][ii];

      for (int ii=0; ii<nCompo; ii++)
        stressG[ii][ii] += lambda(quadPos) * traceG;

      double energyDensityfG = 0;
      for (int ii=0; ii<nCompo; ii++)
        for (int jj=0; jj<nCompo; jj++)
          energyDensityfG += stressG[ii][jj] * forceTerm(quadPos)[ii][jj];                                          

      elementRhs[localPhiOffset+m] += energyDensityfG * quadPoint.weight() * integrationElement;
    }

  }
}

///////////////////////////////////////////////// TEST ///////////////////////////////////////





/*

// Compute the source term for a single element
// < L (sym[D_gamma*nabla phi_i] + M_i ), x_3G_alpha >
template<class LocalView, class LocalFunction1, class LocalFunction2, class Vector, class Force>
void computeElementLoadTEST( const LocalView& localView,
                               LocalFunction1& mu,
                               LocalFunction2& lambda,
                               const double gamma,
                               Vector& elementRhs,
                               const Force& forceTerm
                               )
{
  // | f*phi|
  // | ---  |
  // | f*m  |

  using Element = typename LocalView::Element;
  const auto element = localView.element();
  const auto geometry = element.geometry();
  constexpr int dim = Element::dimension;
  constexpr int nCompo = dim;

  using VectorRT = FieldVector< double, nCompo>;
  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;

  // Set of shape functions for a single element
  const auto& localFiniteElement= localView.tree().child(0).finiteElement();
  const auto nSf = localFiniteElement.localBasis().size();
  //     const auto& localFiniteElement = localView.tree().finiteElement();


  elementRhs.resize(localView.size() +3);
  elementRhs = 0;


  // LocalBasis-Offset
  const int localPhiOffset = localView.size();

  ///////////////////////////////////////////////
  // Basis for R_sym^{2x2}
  //////////////////////////////////////////////
  MatrixRT G_1 {{1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0, 0.0}};
  MatrixRT G_2 {{0.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0, 0.0, 0.0}};
  MatrixRT G_3 {{0.0, 1.0/sqrt(2), 0.0}, {1.0/sqrt(2), 0.0, 0.0}, {0.0, 0.0, 0.0}};
  std::array<MatrixRT,3 > basisContainer = {G_1, G_2, G_3};


  int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);   // für simplex
  const auto& quad = QuadratureRules<double,dim>::rule(element.type(), orderQR);


  for (const auto& quadPoint : quad) 
  {


    const FieldVector<double,dim>& quadPos = quadPoint.position();
    const auto jacobian = geometry.jacobianInverseTransposed(quadPos);
    const double integrationElement = geometry.integrationElement(quadPos);

    std::vector<FieldMatrix<double,1,nCompo> > referenceGradients;
    localFiniteElement.localBasis().evaluateJacobian(quadPos, referenceGradients);

    std::vector< VectorRT > gradients(referenceGradients.size());
    for (size_t i=0; i< gradients.size(); i++)
      jacobian.mv(referenceGradients[i][0], gradients[i]);

    
    auto strain1 = forceTerm(quadPos);

    // "f*phi"-part
    for (size_t i=0; i < nSf; i++)
      for (size_t k=0; k < nCompo; k++)
      {
        // Deformation gradient of the ansatz basis function
        MatrixRT defGradientV(0);
        defGradientV[k][0] = gradients[i][0];                                 // Y
        defGradientV[k][1] = gradients[i][1];                                 // X2
        defGradientV[k][2] = (1.0/gamma)*gradients[i][2];                     // X3

        // Elastic Strain
        MatrixRT strainV;                                                       //strainV never used?
        for (int ii=0; ii<nCompo; ii++)
          for (int jj=0; jj<nCompo; jj++)
            strainV[ii][jj] = 0.5 * (defGradientV[ii][jj] + defGradientV[jj][ii]);

          
          
        
        
        
        // St.Venant-Kirchhoff stress
        MatrixRT stressV(0);
        stressV.axpy(2*mu(quadPos), strain1);                 //this += 2mu *strainU

        double trace = 0.0;
        for (int ii=0; ii<nCompo; ii++)
          trace += strain1[ii][ii];

        for (int ii=0; ii<nCompo; ii++)
          stressV[ii][ii] += lambda(quadPos) * trace;


        // Local energy density: stress times strain
        double energyDensity = 0.0;
        for (int ii=0; ii<nCompo; ii++)
          for (int jj=0; jj<nCompo; jj++)
            energyDensity += stressV[ii][jj] * strainV[ii][jj];


        size_t row = localView.tree().child(k).localIndex(i);
        elementRhs[row] += energyDensity * quadPoint.weight() * integrationElement;
      }


    // "f*m"-part
    for (size_t m=0; m<3; m++)
    {

      // St.Venant-Kirchhoff stress
      MatrixRT stressG(0);
      stressG.axpy(2*mu(quadPos), strain1);       //this += 2mu *strainU

      double traceG = 0;
      for (int ii=0; ii<nCompo; ii++)
        traceG += strain1[ii][ii];

      for (int ii=0; ii<nCompo; ii++)
        stressG[ii][ii] += lambda(quadPos) * traceG;

      double energyDensityfG = 0;
      for (int ii=0; ii<nCompo; ii++)
        for (int jj=0; jj<nCompo; jj++)
          energyDensityfG += stressG[ii][jj] * basisContainer[m][ii][jj];

      elementRhs[localPhiOffset+m] += energyDensityfG * quadPoint.weight() * integrationElement;
    }

  }
}





*/















////////////////////////////////////////////////////////////////////////////////////////////////




template<class Basis, class Matrix, class LocalFunction1, class LocalFunction2>
void assembleCellStiffness(const Basis& basis,
                           LocalFunction1& muLocal,
                           LocalFunction2& lambdaLocal,
                           const double gamma,
                           Matrix& matrix
                           )
{
  std::cout << "assemble stiffnessmatrix begins." << std::endl;

  MatrixIndexSet occupationPattern;
  getOccupationPattern(basis, occupationPattern);
  occupationPattern.exportIdx(matrix);
  matrix = 0;

  
  
  std::fstream localdebugLog;
  localdebugLog.open("../../outputs/debugLog.txt",std::ios::out);


  auto localView = basis.localView();
  const int phiOffset = basis.dimension();


  for (const auto& element : elements(basis.gridView()))
  {
    localView.bind(element);
    muLocal.bind(element);
    lambdaLocal.bind(element);


    const int localPhiOffset = localView.size();
    //         std::cout << "localPhiOffset : " << localPhiOffset << std::endl;

    //         Dune::Matrix<double> elementMatrix;
    Dune::Matrix<FieldMatrix<double,1,1> > elementMatrix;
    computeElementStiffnessMatrix(localView, elementMatrix, muLocal, lambdaLocal, gamma);
//     printmatrix(std::cout, elementMatrix, "ElementMatrix", "--");
//     localdebugLog <<  elementMatrix << std::endl;


    //         std::cout << "elementMatrix.N() : " << elementMatrix.N() << std::endl;
    //         std::cout << "elementMatrix.M() : " << elementMatrix.M() << std::endl;



    //////////////////////////////////////////////////////////////////////////////
    // GLOBAL STIFFNES ASSEMBLY
    //////////////////////////////////////////////////////////////////////////////
    for (size_t i=0; i<localPhiOffset; i++)
    {

      auto row = localView.index(i);

      for (size_t j=0; j<localPhiOffset; j++ )
      {

        auto col = localView.index(j);

        matrix[row][col] += elementMatrix[i][j];
      }
    }

    for (size_t i=0; i<localPhiOffset; i++)
      for(size_t m=0; m<3; m++)
      {
        auto row = localView.index(i);

        matrix[row][phiOffset+m] += elementMatrix[i][localPhiOffset+m];
        matrix[phiOffset+m][row] += elementMatrix[localPhiOffset+m][i];
      }


    for (size_t m=0; m<3; m++ )
      for (size_t n=0; n<3; n++ )
      {
        matrix[phiOffset+m][phiOffset+n] += elementMatrix[localPhiOffset+m][localPhiOffset+n];
      }

//     printmatrix(std::cout, matrix, "StiffnessMatrix", "--");

  }


}


template<class Basis, class LocalFunction1, class LocalFunction2, class Vector, class Force>
void assembleCellLoad(const Basis& basis,
                      LocalFunction1& muLocal,
                      LocalFunction2& lambdaLocal,
                      const double gamma,
                      Vector& b,
                      const Force& forceTerm
                      )
{

  //     std::cout << "assemble load vector." << std::endl;

  b.resize(basis.size()+3);
  b = 0;

  auto localView = basis.localView();

  const int phiOffset = basis.dimension();

  // Transform G_alpha's to GridViewFunctions/LocalFunctions  -- why exactly?
  auto loadGVF  = Dune::Functions::makeGridViewFunction(forceTerm, basis.gridView());
  auto loadFunctional = localFunction(loadGVF);        // Dune::Functions:: notwendig?

  for (const auto& element : elements(basis.gridView()))
  {

    localView.bind(element);
    muLocal.bind(element);
    lambdaLocal.bind(element);
    loadFunctional.bind(element);

    const int localPhiOffset = localView.size();
    //         std::cout << "localPhiOffset : " << localPhiOffset << std::endl;


    //         BlockVector<double> elementRhs;
    BlockVector<FieldVector<double,1> > elementRhs;
    computeElementLoadVector(localView, muLocal, lambdaLocal, gamma, elementRhs, loadFunctional);
//     printvector(std::cout, elementRhs, "elementRhs", "--");
//     computeElementLoadTEST(localView, muLocal, lambdaLocal, gamma, elementRhs, loadFunctional);
//     printvector(std::cout, elementRhs, "elementRhs", "--");

    // LOAD ASSEMBLY
    for (size_t p=0; p<localPhiOffset; p++)
    {
      // The global index of the p-th vertex of the element
      auto row = localView.index(p);

      b[row] += elementRhs[p];
    }

    for (size_t m=0; m<3; m++ )
    {
      b[phiOffset+m] += elementRhs[localPhiOffset+m];
    }
    
    
//     printvector(std::cout, b, "b", "--");
  }

}











template<class Basis, class LocalFunction1, class LocalFunction2, class MatrixFunction>
auto energy(const Basis& basis,
            LocalFunction1& muLocal,
            LocalFunction2& lambdaLocal,
            const MatrixFunction& matrixFieldFuncA,
            const MatrixFunction& matrixFieldFuncB)
{

  double energy = 0.0;
  constexpr int dim = 3;
  constexpr int nCompo = 3;

  auto localView = basis.localView();

  auto matrixFieldAGVF  = Dune::Functions::makeGridViewFunction(matrixFieldFuncA, basis.gridView());
  auto matrixFieldA = localFunction(matrixFieldAGVF);
  auto matrixFieldBGVF  = Dune::Functions::makeGridViewFunction(matrixFieldFuncB, basis.gridView());
  auto matrixFieldB = localFunction(matrixFieldBGVF);

  using GridView = typename Basis::GridView;
  using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;
  
  
  
  // TEST 
//   FieldVector<double,3> testvector = {1.0 , 1.0 , 1.0};
//   printmatrix(std::cout, matrixFieldFuncB(testvector) , "matrixFieldB(testvector) ", "--");



  for (const auto& e : elements(basis.gridView()))
  {
    localView.bind(e);
    matrixFieldA.bind(e);
    matrixFieldB.bind(e);
    muLocal.bind(e);
    lambdaLocal.bind(e);


    FieldVector<double,1> elementEnergy(0);

    auto geometry = e.geometry();
    const auto& localFiniteElement = localView.tree().child(0).finiteElement();

//     int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1)+3;  // TEST 
    int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);
    const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(e.type(), orderQR);

    for (const auto& quadPoint : quad) {
//     for (size_t pt=0; pt < quad.size(); pt++) {
        
        
      const FieldVector<double,dim>& quadPos = quadPoint.position();
//       const Domain& quadPos = quad[pt].position();
      const double integrationElement = geometry.integrationElement(quadPos);


      auto strain1 = matrixFieldA(quadPos);

      // St.Venant-Kirchhoff stress
      MatrixRT stressU(0);
      stressU.axpy(2*muLocal(quadPos), strain1);                 //this += 2mu *strainU        // eigentlich this += 2mu *strain1 ?

      double trace = 0.0;
      for (int ii=0; ii<nCompo; ii++)
        trace += strain1[ii][ii];

      for (int ii=0; ii<nCompo; ii++)
        stressU[ii][ii] += lambdaLocal(quadPos) * trace;

      auto strain2 = matrixFieldB(quadPos);

      // Local energy density: stress times strain
      double energyDensity = 0.0;
      for (int ii=0; ii<nCompo; ii++)
        for (int jj=0; jj<nCompo; jj++)
          energyDensity += stressU[ii][jj] * strain2[ii][jj];

//       elementEnergy += energyDensity * quad[pt].weight() * integrationElement;
      elementEnergy += energyDensity * quadPoint.weight() * integrationElement;
    }
    energy += elementEnergy;
  }
  return energy;
}






























template<class Basis, class Matrix, class Vector>
void setOneBaseFunctionToZero(const Basis& basis,
                              Matrix& stiffnessMatrix,
                              Vector& load_alpha1,
                              Vector& load_alpha2,
                              Vector& load_alpha3,
                              FieldVector<int,3>& indexVector  // oder input : arbitraryLocalIndex und dann hier function computeArbitraryIndex aufrufen
                              )
{

  std::cout << "setting one Basis-function to zero" << std::endl;

  constexpr int dim = 3;
  auto localView = basis.localView();



  //Determine 3 global indices (one for each component of an arbitrary FE-function)..    oder mittels Offset?

  // use first element:
  auto it = basis.gridView().template begin<0>();
  localView.bind(*it);

  int arbitraryIndex = 0;
  FieldVector<int,3> row;        //fill with Indices..

  for (int k = 0; k < dim; k++)
  {
    auto rowLocal = localView.tree().child(k).localIndex(arbitraryIndex);
    row[k] = localView.index(rowLocal);
    //         std::cout << "row[k]:" << row[k] << std::endl;
    load_alpha1[row[k]] = 0.0;                                                              //verwende hier indexVector
    load_alpha2[row[k]] = 0.0;
    load_alpha3[row[k]] = 0.0;
    auto cIt    = stiffnessMatrix[row[k]].begin();
    auto cEndIt = stiffnessMatrix[row[k]].end();
    for (; cIt!=cEndIt; ++cIt)
      *cIt = (cIt.index()==row[k]) ? 1.0 : 0.0;
  }

}






template<class Basis>
auto childToIndexMap(const Basis& basis,
                     const int k
                     )
{
  constexpr int dim = 3;



  using GridView = typename Basis::GridView;
  using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;


  using FuncScalar = std::function< FieldVector< double, dim>(const Domain&) >;

  auto localView = basis.localView();



  std::vector<int> r = { };
  //     for (int n: r)
  //         std::cout << n << ","<< std::endl;

  // Determine global indizes for each component k = 1,2,3.. in order to subtract correct (component of) integral Mean
  // (global) Indices that correspond to component k = 1,2,3
  for(const auto& element : elements(basis.gridView()))
  {

    localView.bind(element);
    const auto& localFiniteElement = localView.tree().child(k).finiteElement();        // macht keinen Unterschied ob hier k oder 0..
    const auto nSf = localFiniteElement.localBasis().size();


    for(size_t j=0; j<nSf; j++)
    {
      auto Localidx = localView.tree().child(k).localIndex(j);  // local  indices 
      r.push_back(localView.index(Localidx));                   // global indices
    }



  }

  // Delete duplicate elements
  // first remove consecutive (adjacent) duplicates
  auto last = std::unique(r.begin(), r.end());
  r.erase(last, r.end());
  // sort followed by unique, to remove all duplicates
  std::sort(r.begin(), r.end());
  last = std::unique(r.begin(), r.end());
  r.erase(last, r.end());


  return r;


}







template<class Basis, class Vector>
auto integralMean(const Basis& basis,
                  Vector& coeffVector
                  )
{
  // computes integral mean of given LocalFunction
    
  
  


  using GridView = typename Basis::GridView;
  using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
  constexpr int dim = 3;
  using FuncScalar = std::function< FieldVector< double, dim>(const Domain&) >;
  
  
  auto GVFunction = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,dim>>(basis,coeffVector);
  auto localfun = localFunction(GVFunction);
  
  

  auto localView = basis.localView();
  //     double r = 0.0;
  FieldVector<double,3> r = {0.0, 0.0, 0.0};
  double area = 0.0;

  // Compute area integral
  for(const auto& element : elements(basis.gridView()))
  {

    localView.bind(element);
    const auto& localFiniteElement = localView.tree().child(0).finiteElement();
    
//     int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1)+5;  // TEST 
    int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);
    const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(element.type(), orderQR);

    for (const auto& quadPoint : quad)
    {
      const double integrationElement = element.geometry().integrationElement(quadPoint.position());
      area += quadPoint.weight() * integrationElement;
    }
  }
  //     std::cout << "Domain-Area: " << area << std::endl;


  for(const auto& element : elements(basis.gridView()))
  {

    localView.bind(element);
    localfun.bind(element);
    const auto& localFiniteElement = localView.tree().child(0).finiteElement();      
    int order = 2*(dim*localFiniteElement.localBasis().order()-1);
    const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(element.type(), order);

    for (const auto& quadPoint : quad)
    {

      const auto& quadPos = quadPoint.position();
      const double integrationElement = element.geometry().integrationElement(quadPos);


      //         std::vector<FieldMatrix<double,1,nCompo> > referenceGradients;
      //        localFiniteElement.localBasis().evaluateJacobian(quadPos, referenceGradients);
      //
      //        std::vector< VectorRT > gradients(referenceGradients.size());
      //        for (size_t i=0; i< gradients.size(); i++)
      //                jacobian.mv(referenceGradients[i][0], gradients[i]);

      auto value = localfun(quadPos);
      //         auto value = localPhi_1(quadPos);
      //         auto value = FieldVector<double,dim>{1.0, 1.0, 1.0};

      //         for(size_t k=0; k<3; k++)
      //             r += value[k] * quadPoint.weight() * integrationElement;
      r += value * quadPoint.weight() * integrationElement;


      //         r += tmp * quadPoint.weight() * integrationElement;

    }
  }

  //     return r/area;
  return (1.0/area) * r ;

}








template<class Basis, class Vector>
auto subtractIntegralMean(const Basis& basis,
                          Vector& coeffVector
                          )
{
  // subtract correct Integral mean from each associated component function

//   auto IM = integralMean(basis, fun);
  auto IM = integralMean(basis, coeffVector);
  constexpr int dim = 3;

  for(size_t k=0; k<dim; k++)
  {
    //         std::cout << "Integral-Mean: " << IM[k] << std::endl;
    auto idx = childToIndexMap(basis,k);

    for ( int i : idx)
      coeffVector[i] -= IM[k];
  }
}



  //////////////////////////////////////////////////
  //   Infrastructure for handling periodicity
  //////////////////////////////////////////////////

// Check whether two points are equal on R/Z x R/Z x R
auto equivalent = [](const FieldVector<double,3>& x, const FieldVector<double,3>& y)
                  {
//                     std::cout <<  ( (FloatCmp::eq(x[0],y[0]) or FloatCmp::eq(x[0]+1,y[0]) or FloatCmp::eq(x[0]-1,y[0]))
//                              and (FloatCmp::eq(x[1],y[1]) or FloatCmp::eq(x[1]+1,y[1]) or FloatCmp::eq(x[1]-1,y[1]))
//                              and (FloatCmp::eq(x[2],y[2])) ) << std::endl;
                    
                    return ( (FloatCmp::eq(x[0],y[0]) or FloatCmp::eq(x[0]+1,y[0]) or FloatCmp::eq(x[0]-1,y[0]))
                             and (FloatCmp::eq(x[1],y[1]) or FloatCmp::eq(x[1]+1,y[1]) or FloatCmp::eq(x[1]-1,y[1]))
                             and (FloatCmp::eq(x[2],y[2]))
                           );
                  };








// ---- TODO - MatrixEmbedding function (iota) ---- ?
                  
                  
                  
                  
                  
                  

template<class Basis, class Vector, class MatrixFunction, class Domain>        //TODO
double evalSymGrad(const Basis& basis,
                   Vector& coeffVector,
                   const double gamma,
                   Domain& x    // evaluation Point in reference coordinates
                      )
{
  

  constexpr int dim = 3;
  constexpr int nCompo = 3;

  auto localView = basis.localView();


  using GridView = typename Basis::GridView;
//   using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;



  
  
  for (const auto& element : elements(basis.gridView()))
  {
      
    localView.bind(element);
    auto geometry = element.geometry();
    
    const auto& localFiniteElement = localView.tree().child(0).finiteElement();                // Unterscheidung children notwendig?
    const auto nSf = localFiniteElement.localBasis().size();
    


        const auto jacobianInverseTransposed = geometry.jacobianInverseTransposed(x);
  

        std::vector< FieldMatrix< double, 1, dim> > referenceGradients;
        localFiniteElement.localBasis().evaluateJacobian(x,
                                                         referenceGradients);

        // Compute the shape function gradients on the grid element
        std::vector<FieldVector<double,dim> > gradients(referenceGradients.size());
        //         std::vector< VectorRT> gradients(referenceGradients.size());

        for (size_t i=0; i<gradients.size(); i++)
          jacobianInverseTransposed.mv(referenceGradients[i][0], gradients[i]);


        MatrixRT defGradientU(0);
        
        
        for (size_t k=0; k < nCompo; k++)
            for (size_t i=0; i < nSf; i++)          //Error: these are local fcts!
            {
                
                size_t localIdx = localView.tree().child(k).localIndex(i); 
                size_t globalIdx = localView.index(localIdx);

                // (scaled) Deformation gradient of the ansatz basis function
                defGradientU[k][0] += coeffVector[globalIdx]* gradients[i][0];                       // Y
                defGradientU[k][1] += coeffVector[globalIdx]* gradients[i][1];                       //X2
                defGradientU[k][2] += coeffVector[globalIdx]*(1.0/gamma)*gradients[i][2];           //X3
            
                
            
                // symmetric Gradient (Elastic Strains)
                MatrixRT strainU(0);
                for (int ii=0; ii<nCompo; ii++)
                for (int jj=0; jj<nCompo; jj++)
                {
                    strainU[ii][jj] = coeffVector[globalIdx] * 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);                 // symmetric gradient
                }
                
             
            }
  }

}




template<class Basis, class Vector, class MatrixFunction>
double computeL2Error(const Basis& basis,
                      Vector& coeffVector,
                      const double gamma,
                      const MatrixFunction& matrixFieldFunc
                      )
{
  
  auto error = 0.0;
  constexpr int dim = 3;
  constexpr int nCompo = 3;

  auto localView = basis.localView();

  auto matrixFieldGVF  = Dune::Functions::makeGridViewFunction(matrixFieldFunc, basis.gridView());
  auto matrixField = localFunction(matrixFieldGVF);

  using GridView = typename Basis::GridView;
  using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;



  
  
  for (const auto& element : elements(basis.gridView()))
  {
      
    localView.bind(element);
    matrixField.bind(element);
    auto geometry = element.geometry();
    
    const auto& localFiniteElement = localView.tree().child(0).finiteElement();                // Unterscheidung children notwendig?
    const auto nSf = localFiniteElement.localBasis().size();
    

//     int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1)+5;  // TEST 
    int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);
    const auto& quad = QuadratureRules<double,dim>::rule(element.type(), orderQR);


    
        
    for (const auto& quadPoint : quad)
    {

        const auto& quadPos = quadPoint.position();
        const auto jacobianInverseTransposed = geometry.jacobianInverseTransposed(quadPoint.position());
        const auto integrationElement = geometry.integrationElement(quadPoint.position());

        std::vector< FieldMatrix< double, 1, dim> > referenceGradients;
        localFiniteElement.localBasis().evaluateJacobian(quadPoint.position(),
                                                         referenceGradients);

        // Compute the shape function gradients on the grid element
        std::vector<FieldVector<double,dim> > gradients(referenceGradients.size());
        //         std::vector< VectorRT> gradients(referenceGradients.size());

        for (size_t i=0; i<gradients.size(); i++)
          jacobianInverseTransposed.mv(referenceGradients[i][0], gradients[i]);


        MatrixRT defGradientU(0);
        
        
        for (size_t k=0; k < nCompo; k++)
            for (size_t i=0; i < nSf; i++)
            {
                
                size_t localIdx = localView.tree().child(k).localIndex(i); 
                size_t globalIdx = localView.index(localIdx);

                // (scaled) Deformation gradient of the ansatz basis function
                defGradientU[k][0] = gradients[i][0];                       // Y
                defGradientU[k][1] = gradients[i][1];                       //X2
                defGradientU[k][2] = (1.0/gamma)*gradients[i][2];           //X3
            
                
            
                // symmetric Gradient (Elastic Strains)
                MatrixRT strainU(0);
                for (int ii=0; ii<nCompo; ii++)
                for (int jj=0; jj<nCompo; jj++)
                {
                    strainU[ii][jj] = coeffVector[globalIdx] * 0.5 * (defGradientU[ii][jj] + defGradientU[jj][ii]);                 // symmetric gradient
                }
                
             

                // Local energy density: stress times strain
                double tmp = 0.0;
                for (int ii=0; ii<nCompo; ii++)
                for (int jj=0; jj<nCompo; jj++)
                    tmp+= std::pow(strainU[ii][jj] - matrixField(quadPos)[ii][jj],2);
//                     tmp+= std::pow((coeffVector[globalIdx]*strainU[ii][jj]) - matrixField(quadPos)[ii][jj] ,2);

        
                
                error += tmp * quadPoint.weight() * integrationElement;
            }
    }
  }
  
  return sqrt(error);
  
}








int main(int argc, char *argv[])
{

  MPIHelper::instance(argc, argv);



  ParameterTree parameterSet;
  //     if (argc < 2)
  //     ParameterTreeParser::readINITree("../src/cellsolver.parset", parameterSet);
  //     ParameterTreeParser::readINITree(argv[1], parameterSet);
  //     ParameterTreeParser::readOptions(argc, argv, parameterSet);
  
  // output setter
  // -- not set up --
  std::fstream log;
  std::fstream debugLog;
//   log.open("output2.txt", std::ios::out);
  log.open("../../outputs/output.txt",std::ios::out);
  debugLog.open("../../outputs/debugLog.txt",std::ios::out);
//   log << "Writing this to a file. \n";
//   log.close();
  

  constexpr int dim = 3;
  constexpr int nCompo = 3;
  constexpr int order_CE = 1;

  ///////////////////////////////////
  // Get Parameters/Data
  ///////////////////////////////////
  double gamma = parameterSet.get<double>("gamma",2000.0);   // ratio dimension reduction to homogenization

  // Material parameter laminat
  //     double E1                = parameterSet.get<double>("E1", 17e6); //material1
  //     double nu1               = parameterSet.get<double>("nu1", 0.0);
  //     double nu1               = parameterSet.get<double>("nu1", 0.3);
  //     double E2                = parameterSet.get<double>("E2", 35e6); //material2
  //     double nu2               = parameterSet.get<double>("nu2", 0.5);

  //     // Plate geometry settings
  double width = parameterSet.get<double>("width", 1.0);   //geometry cell, cross section
  //     double len  = parameterSet.get<double>("len", 10.0); //length
  //     double height  = parameterSet.get<double>("h", 0.1); //rod thickness
  //     double eps  = parameterSet.get<double>("eps", 0.1); //size of perioticity cell


  ///////////////////////////////////
  // ---Get Prestrain ---
  ///////////////////////////////////
  unsigned int prestraintype = parameterSet.get<unsigned int>("imp", 1);
  double p1 = parameterSet.get<double>("prestress_pressure1", 1.0);
  double p2 = parameterSet.get<double>("prestress_pressure2", 2.0);
  double theta = parameterSet.get<double>("theta",0.3);
  //     double theta = 0.5;
  p1 = 1.0;
  double alpha = 2;
  p2 = alpha*1.0;
  
  prestraintype = 2;
  //     prestraintype = 1;

  auto prestrainImp = PrestrainImp(p1, p2, theta, width);
  auto B_Term = prestrainImp.getPrestrain(prestraintype);

  log << "prestrain imp: " <<  prestraintype << "\np1 = " << p1 << "\np2 = " << p2  << std::endl;
  log << "alpha: " << alpha << std::endl;
  log << "gamma: " << gamma << std::endl; 
  
  
  ///////////////////////////////////
  // --- Choose Solver ---
  // 1 : CG-Solver
  // 2 : GMRES
  // 3 : QR 
  ///////////////////////////////////
  unsigned int Solvertype = parameterSet.get<unsigned int>("Solvertype", 1);
  bool write_corrector_phi1 = parameterSet.get<bool>("write_corrector_phi1", true);
  bool write_corrector_phi2 = parameterSet.get<bool>("write_corrector_phi2", true);
  bool write_corrector_phi3 = parameterSet.get<bool>("write_corrector_phi3", true);
  
  ///////////////////////////////////
  // Generate the grid
  ///////////////////////////////////
  using CellGridType = YaspGrid<dim, EquidistantOffsetCoordinates<double, dim> >;

  // Domain : [0,1)^2 x (-1/2, 1/2)
  //     FieldVector<double,dim> lower({0.0, 0.0, -1.0/2.0});
  //     FieldVector<double,dim> upper({1.0, 1.0, 1.0/2.0});

  // (shifted) Domain (-1/2,1/2)^3
  FieldVector<double,dim> lower({-1.0/2.0, -1.0/2.0, -1.0/2.0});
  FieldVector<double,dim> upper({1.0/2.0, 1.0/2.0, 1.0/2.0});

  int nE = 10;
  log << "Number of Elements: " << nE << std::endl;
  debugLog << "Number of Elements: " << nE << std::endl;
  
  std::array<int,dim> nElements={nE,nE,nE};
//   std::array<int,dim> nElements={100,100,1};  // TEST
//   printvector(std::cout, coeffT_1  , "coeffT_1", "--" );
//   printvector(std::cout,  x_1 , "x_1", "--" );;    //#Elements in each direction

  CellGridType grid_CE(lower,upper,nElements);   //Corrector Problem Domain:

  using GridView = CellGridType::LeafGridView;
  const GridView gridView_CE = grid_CE.leafGridView();



  using ScalarRT = FieldVector< double, 1>;
  using VectorRT = FieldVector< double, nCompo>;
  using MatrixRT = FieldMatrix< double, nCompo, nCompo>;
  using Domain = GridView::Codim<0>::Geometry::GlobalCoordinate;
  using FuncScalar = std::function< ScalarRT(const Domain&) >;
  using Func2Tensor = std::function< MatrixRT(const Domain&) >;

  using VectorCT = BlockVector<FieldVector<double,1> >;
  using MatrixCT = BCRSMatrix<FieldMatrix<double,1,1> >;
  // using VectorCT = BlockVector<FieldVector<double,dim> >;
  // using MatrixCT = BCRSMatrix<FieldMatrix<double,dim,dim> >;
  using ElementMatrixCT = Matrix<FieldMatrix<double,1,1> >;





//   double beta = 1;   //homogeneous case
  double beta = 2;
  
  
  double mu1 = 10;
//   double mu1 = 0.5*17e6;
  
  //     double mu1 = 1000;
  double mu2 = beta*mu1;
  // //

  //     double mu1 = 0.5 * 1.0/(1.0 + nu1) * E1;
  //     double lambda1 = nu1/(1.0-2.0*nu1) * 1.0/(1.0+nu1) * E1;
  double lambda1 = 0;


  // Create Lambda-Functions for material Parameters
  //     auto muTerm = [mu1] (const Domain& z) {
  //         return mu1;
  //     };

  auto muTerm = [mu1, mu2, theta] (const Domain& z) {
                  if (abs(z[0]) >= (theta/2))
                    return mu1;
                  //         if (abs(z[0]) < (theta/2) && z[2] < 0)   // oder das?
                  //             return mu2;
                  else
                    return mu2;
                  //             return mu1;
                };





  auto lambdaTerm = [lambda1] (const Domain& z) {
                      return lambda1;
                    };


  auto muGridF  = Dune::Functions::makeGridViewFunction(muTerm, gridView_CE);
  auto muLocal = localFunction(muGridF);
  auto lambdaGridF  = Dune::Functions::makeGridViewFunction(lambdaTerm, gridView_CE);
  auto lambdaLocal = localFunction(lambdaGridF);
  
  log << "beta: " << beta << std::endl;
  log << "material parameters: " << std::endl; 
  log << " mu1: " << mu1 << "\nmu2: " << mu2 << std::endl;
  log << "\n lambda: " << lambda1 << std::endl;
  
  
  
  /////////////////////////////////////////////////////////
  // Choose a finite element space for Cell Problem
  /////////////////////////////////////////////////////////
  using namespace Functions::BasisFactory;

  Functions::Experimental::BasisFactory::PeriodicIndexSet periodicIndices;

  int equivCounter = 0;

  
  // Don't do the following in real life: It has quadratic run-time in the number of vertices.
  for (const auto& v1 : vertices(gridView_CE))
  {
//     std::cout << " ---------------------------------------" << std::endl; 
    for (const auto& v2 : vertices(gridView_CE))
      if (equivalent(v1.geometry().corner(0), v2.geometry().corner(0)))
      {
        periodicIndices.unifyIndexPair({gridView_CE.indexSet().index(v1)}, {gridView_CE.indexSet().index(v2)});
        equivCounter++;
      }
  }
  std::cout << "equivCounter:" <<  equivCounter << std::endl;
  std::cout << "Number of Elements in each direction: " << nE << std::endl;
  std::cout << "Number ofNodes: " << gridView_CE.size(dim) << std::endl;
  
  auto perTest = periodicIndices.indexPairSet();
  
  
  auto Basis_CE = makeBasis(
    gridView_CE,
    power<dim>(       //lagrange<1>(),
      Functions::Experimental::BasisFactory::periodic(lagrange<1>(), periodicIndices),
      flatLexicographic()));      //
//         blockedInterleaved()));    // siehe Periodicbasistest.. funktioniert auch?
//             flatInterleaved()));


  std::cout << "size feBasis: " << Basis_CE.size() << std::endl;
  std::cout << "basis.dimension()" << Basis_CE.dimension() <<std::endl;

  const int phiOffset = Basis_CE.size();
  debugLog << "phiOffset: "<< phiOffset << std::endl;
  


//   std::cout << Basis_CE.preBasis_.children << std::endl;   // does not work
//   std::cout << Basis_CE::preBasis_::children << std::endl;
  // TEST
  //     auto X1Basis = subspaceBasis(Basis_CE, 0);
  //     auto X2Basis = subsppaceBasis(Basis_CE, 1);
  //     auto X3Basis = subspaceBasis(Basis_CE, 2);
  //
  //     std::cout << "size X1Basis: " << X1Basis.size() << std::endl;
  //     std::cout << "X1Basis.dimension()" << X1Basis.dimension() <<std::endl;
  //     std::cout << "size X2Basis: " << X2Basis.size() << std::endl;
  //     std::cout << "size X3Basis: " << X3Basis.size() << std::endl;



  /////////////////////////////////////////////////////////
  // Stiffness matrix and right hand side vector
  /////////////////////////////////////////////////////////
  VectorCT load_alpha1, load_alpha2, load_alpha3;
  MatrixCT stiffnessMatrix_CE;
  // auto rhsBackend = Functions::istlVectorBackend(b);
  // rhsBackend.resize(Basis_CE);
  // b = 0;



  bool set_IntegralZero = parameterSet.get<bool>("set_IntegralZero", true);
  bool set_oneBasisFunction_Zero = false;
  bool substract_integralMean = false;
  if(set_IntegralZero)
  {
    set_oneBasisFunction_Zero = true;
    substract_integralMean = true;
  }

  debugLog << " set_oneBasisFunction_Zero: " <<  set_oneBasisFunction_Zero << std::endl;
  debugLog << " substract_integralMean: " << substract_integralMean << std::endl;



  // using Range = FieldVector<double,dim>;
  // auto sourceTerm = [](const FieldVector<double,dim>& x){return Range{0.0, -1.0};};

  Func2Tensor x3G_1 = [] (const Domain& x) {
                        return MatrixRT{{1.0*x[2], 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
                      };

  Func2Tensor x3G_2 = [] (const Domain& x) {
                        return MatrixRT{{0.0, 0.0, 0.0}, {0.0, 1.0*x[2], 0.0}, {0.0, 0.0, 0.0}};
                      };

  Func2Tensor x3G_3 = [] (const Domain& x) {
                        return MatrixRT{{0.0, 1.0/sqrt(2)*x[2], 0.0}, {1.0/sqrt(2)*x[2], 0.0, 0.0}, {0.0, 0.0, 0.0}}; 
                      };

                      
                      
  Func2Tensor x3G_1neg = [x3G_1] (const Domain& x) {
                        return -1.0*x3G_1(x);
                      };

  Func2Tensor x3G_2neg = [x3G_2] (const Domain& x) {
                        return -1.0*x3G_2(x);;
                      };

  Func2Tensor x3G_3neg = [x3G_3] (const Domain& x) {
                        return -1.0*x3G_3(x);
                      };
//     FieldVector<double,3> test= {1.0/4.0 , 0.0 , 0.25};
//     auto Tast = x3G_3(test);
//     printmatrix(std::cout, Tast, "x3G_3", "--");

  ///////////////////////////////////////////////////////////////////////                         
  // double theta = 0.5;
  // double p1 = 1;
  // double p2 = 2;

//   using std::abs;


//   Func2Tensor B = [] (const Domain& x) {
//                     double theta = 0.5;
//                     double p1 = 1;
//                     double p2 = 2;
//                     if (abs(x[0])>= (theta/2)  && x[2] > 0)
//                       return MatrixRT{{p1, 0.0 , 0.0}, {0.0, p1, 0.0}, {0.0, 0.0, p1}};
//                     if (abs(x[0]) < (theta/2)  && x[2] < 0)
//                       return MatrixRT{{p2, 0.0 , 0.0}, {0.0, p2, 0.0}, {0.0, 0.0, p2}};
//                     else
//                       return MatrixRT{{0.0, 0.0 , 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
// 
//                   };
  //////////////////////////////////////////////////////////////////////////////


  ///////////////////////////////////////////////
  // Basis for R_sym^{2x2}
  //////////////////////////////////////////////
  MatrixRT G_1 {{1.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0, 0.0}};
  MatrixRT G_2 {{0.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0, 0.0, 0.0}};
  MatrixRT G_3 {{0.0, 1.0/sqrt(2), 0.0}, {1.0/sqrt(2), 0.0, 0.0}, {0.0, 0.0, 0.0}}; 

  std::array<MatrixRT,3 > basisContainer = {G_1, G_2, G_3};
//   printmatrix(std::cout, basisContainer[0] , "G_1", "--");
  // printmatrix(std::cout, basisContainer[1] , "G_2", "--");
  // printmatrix(std::cout, basisContainer[2] , "´G_3", "--");
//   log <<  basisContainer[0] << std::endl;



  /////////////////////////////////////////////////////////
  // Assemble the system
  /////////////////////////////////////////////////////////
  assembleCellStiffness(Basis_CE, muLocal, lambdaLocal, gamma,  stiffnessMatrix_CE);
  assembleCellLoad(Basis_CE, muLocal, lambdaLocal,gamma, load_alpha1 ,x3G_1neg);
  assembleCellLoad(Basis_CE, muLocal, lambdaLocal,gamma, load_alpha2 ,x3G_2neg);
  assembleCellLoad(Basis_CE, muLocal, lambdaLocal,gamma, load_alpha3 ,x3G_3neg);
  
  
  // TEST multiply RHS with -1 : 
//     for(int i=0; i< Basis_CE.size()+3 ; i++)
//     {
//         load_alpha1[i] = (-1.0)*load_alpha1[i];
//         load_alpha2[i] = (-1.0)*load_alpha2[i];
//         load_alpha3[i] = (-1.0)*load_alpha3[i];
//     }     


  
//   printmatrix(std::cout, stiffnessMatrix_CE, "StiffnessMatrix before B.C", "--");
//   printvector(std::cout, load_alpha1, "load_alpha1", "--");

  
  
  /////////////////////////////////////////////////////////
  // Determine global indices of components for arbitrary (local) index        TODO :separate Function! arbitraryComponentwiseIndices
  /////////////////////////////////////////////////////////
  auto localView = Basis_CE.localView();

  int arbitraryIndex = 0;

  FieldVector<int,3> row;    //fill with Indices..


  // use first element:
  auto it = Basis_CE.gridView().template begin<0>();
  localView.bind(*it);

  for (int k = 0; k < dim; k++)
  {
    auto rowLocal = localView.tree().child(k).localIndex(arbitraryIndex);
    row[k] = localView.index(rowLocal);
    std::cout << "row[k]:" << row[k] << std::endl;

  }
  ///////////////////////////////////////////////////////////
  
  


  // set one basis-function to zero
  if(set_oneBasisFunction_Zero)
  {
    setOneBaseFunctionToZero(Basis_CE, stiffnessMatrix_CE, load_alpha1, load_alpha2, load_alpha3, row);
//     printmatrix(std::cout, stiffnessMatrix_CE, "StiffnessMatrix after setOneBasisFunctionToZero", "--");
//     printvector(std::cout, load_alpha1, "load_alpha1 after setOneBaseFunctionToZero", "--");
  }



  ////////////////////////////////////////////////////
  // Compute solution
  ////////////////////////////////////////////////////
  VectorCT x_1 = load_alpha1;
  VectorCT x_2 = load_alpha2;
  VectorCT x_3 = load_alpha3;
  
  auto load_alpha1BS = load_alpha1;
  
//   printvector(std::cout, load_alpha1, "load_alpha1 before SOLVER", "--" );
//   printvector(std::cout, load_alpha2, "load_alpha2 before SOLVER", "--" );
    
  if (Solvertype == 1)  // CG - SOLVER
  {
    std::cout << "------------ CG - Solver ------------" << std::endl;
    MatrixAdapter<MatrixCT, VectorCT, VectorCT> op(stiffnessMatrix_CE);



    // Sequential incomplete LU decomposition as the preconditioner
    SeqILU<MatrixCT, VectorCT, VectorCT> ilu0(stiffnessMatrix_CE,1.0);
    int iter = parameterSet.get<double>("iterations_CG", 999);
    // Preconditioned conjugate-gradient solver
    CGSolver<VectorCT> solver(op,
                            ilu0, //NULL,
                            1e-8, // desired residual reduction factorlack
                            iter,   // maximum number of iterations
                            2);   // verbosity of the solver
    InverseOperatorResult statistics;
    std::cout << "solve linear system for x_1.\n";
    solver.apply(x_1, load_alpha1, statistics);
    std::cout << "solve linear system for x_2.\n";
    solver.apply(x_2, load_alpha2, statistics);
    std::cout << "solve linear system for x_3.\n";
    solver.apply(x_3, load_alpha3, statistics);
  }
  ////////////////////////////////////////////////////////////////////////////////////
  
  else if (Solvertype ==2)  // GMRES - SOLVER
  {
      
    std::cout << "------------ GMRES - Solver ------------" << std::endl;
    // Turn the matrix into a linear operator
    MatrixAdapter<MatrixCT,VectorCT,VectorCT> stiffnessOperator(stiffnessMatrix_CE);

    // Fancy (but only) way to not have a preconditioner at all
    Richardson<VectorCT,VectorCT> preconditioner(1.0);

    // Construct the iterative solver
    RestartedGMResSolver<VectorCT> solver(
        stiffnessOperator, // Operator to invert
        preconditioner,    // Preconditioner
        1e-10,             // Desired residual reduction factor
        500,               // Number of iterations between restarts,
                        // here: no restarting
        500,               // Maximum number of iterations
        2);                // Verbosity of the solver

    // Object storing some statistics about the solving process
    InverseOperatorResult statistics;

    // solve for different Correctors (alpha = 1,2,3)
    solver.apply(x_1, load_alpha1, statistics);
    solver.apply(x_2, load_alpha2, statistics);
    solver.apply(x_3, load_alpha3, statistics);
  }
  ////////////////////////////////////////////////////////////////////////////////////
  else if ( Solvertype == 3)// QR - SOLVER
  {

    std::cout << "------------ QR - Solver ------------" << std::endl;
    log << "solveLinearSystems: We use QR solver.\n";
    //TODO install suitesparse
    SPQR<MatrixCT> sPQR(stiffnessMatrix_CE);
    sPQR.setVerbosity(1);
    InverseOperatorResult statisticsQR;

    sPQR.apply(x_1, load_alpha1, statisticsQR);
    sPQR.apply(x_2, load_alpha2, statisticsQR);
    sPQR.apply(x_3, load_alpha3, statisticsQR);

  }
//      printvector(std::cout, load_alpha1BS, "load_alpha1 before SOLVER", "--" );
//     printvector(std::cout, load_alpha1, "load_alpha1 AFTER SOLVER", "--" );
    std::cout << "---------load_alpha2 AFTER SOLVER: -------------" << std::endl;
//     printvector(std::cout, load_alpha2, "load_alpha2 AFTER SOLVER", "--" );
        
    
    

  ////////////////////////////////////////////////////////////////////////////////////
  // Extract phi_alpha  & M_alpha coefficients
  ////////////////////////////////////////////////////////////////////////////////////
    
  VectorCT phi_1, phi_2, phi_3;
  phi_1.resize(Basis_CE.size());
  phi_1 = 0;
  phi_2.resize(Basis_CE.size());
  phi_2 = 0;
  phi_3.resize(Basis_CE.size());
  phi_3 = 0;


  for(size_t i=0; i<Basis_CE.size(); i++)
  {
    phi_1[i] = x_1[i];
    phi_2[i] = x_2[i];
    phi_3[i] = x_3[i];
  }
  
  FieldVector<double,3> m_1, m_2, m_3;
  
  for(size_t i=0; i<3; i++)
  {
    m_1[i] = x_1[phiOffset+i];
    m_2[i] = x_2[phiOffset+i];
    m_3[i] = x_3[phiOffset+i];
  }
  
  // assemble M_alpha's (actually iota(M_alpha) )

  MatrixRT M_1(0), M_2(0), M_3(0);

  for(size_t i=0; i<3; i++)
  {
    M_1 += m_1[i]*basisContainer[i];
    M_2 += m_2[i]*basisContainer[i];
    M_3 += m_3[i]*basisContainer[i];
  }

  std::cout << "--- plot corrector-Matrices M_alpha --- " << std::endl;
  printmatrix(std::cout, M_1, "Corrector-Matrix M_1", "--");
  printmatrix(std::cout, M_2, "Corrector-Matrix M_2", "--");
  printmatrix(std::cout, M_3, "Corrector-Matrix M_3", "--");


  
  
  


  ////////////////////////////////////////////////////////////////////////////
  // Substract Integral-mean
  ////////////////////////////////////////////////////////////////////////////                                                                                     // ERROR : 
  using SolutionRange = FieldVector<double,dim>;

  std::cout << "check integralMean phi_1: " << std::endl;
  auto A = integralMean(Basis_CE,phi_1);
  for(size_t i=0; i<3; i++)
  {
    std::cout << "Integral-Mean : " << A[i] << std::endl;
  }
  if(substract_integralMean)
  {
    std::cout << " --- substracting integralMean --- " << std::endl;
    subtractIntegralMean(Basis_CE,  phi_1);
    subtractIntegralMean(Basis_CE,  phi_2);
    subtractIntegralMean(Basis_CE,  phi_3);
    
//     printvector(std::cout,  x_1 , "x_1", "--" );
    subtractIntegralMean(Basis_CE,  x_1);       
    subtractIntegralMean(Basis_CE,  x_2);
    subtractIntegralMean(Basis_CE,  x_3);
//     printvector(std::cout,  x_1 , "x_1 after substract_integralMean", "--" );
    ////////////////////////////////////////////////////////////////////////
    // CHECK INTEGRAL-MEAN:
    auto A = integralMean(Basis_CE, phi_1);
    for(size_t i=0; i<3; i++)
    {
      std::cout << "Integral-Mean (CHECK)  : " << A[i] << std::endl;
    }
    ////////////////////////////////////////////////////
  }

//   // PRINT Corrector-Basis-Coefficients
//   printvector(std::cout, phi_1, "Coefficients Corrector-Phi_1", "--");  
//   printvector(std::cout, phi_2, "Coefficients Corrector-Phi_2", "--");  
//   printvector(std::cout, phi_3, "Coefficients Corrector-Phi_3", "--");  



  /////////////////////////////////////////////////////////
  // Write Solution in Logs
  /////////////////////////////////////////////////////////        
  log << "\nSolution of Corrector problems:\n";
  if(write_corrector_phi1)
  {
     log << "\n Corrector_phi1:\n"; 
     log << x_1 << std::endl;
  }
  if(write_corrector_phi2)
  {
     log << "-----------------------------------------------------";
     log << "\n Corrector_phi2:\n"; 
     log << x_2 << std::endl;
  }
  if(write_corrector_phi3)
  {
     log << "-----------------------------------------------------";
     log << "\n Corrector_phi3:\n"; 
     log << x_3 << std::endl;
  }

  ////////////////////////////////////////////////////////////////////////////
  // Make a discrete function from the FE basis and the coefficient vector
  ////////////////////////////////////////////////////////////////////////////                                                                                     // ERROR : 
  using SolutionRange = FieldVector<double,dim>;
  auto correctorFunction_1 = Functions::makeDiscreteGlobalBasisFunction<SolutionRange>(
    Basis_CE,
    phi_1);

  auto correctorFunction_2 = Functions::makeDiscreteGlobalBasisFunction<SolutionRange>(
    Basis_CE,
    phi_2);

  auto correctorFunction_3 = Functions::makeDiscreteGlobalBasisFunction<SolutionRange>(
    Basis_CE,
    phi_3);

  // Extract phi_alpha  & M_alpha coefficients
  VectorCT coeffT_1,coeffT_2,coeffT_3;
  coeffT_1.resize(Basis_CE.size()+3);
  coeffT_1 = 0;
  coeffT_2.resize(Basis_CE.size()+3);
  coeffT_2 = 0;
  coeffT_3.resize(Basis_CE.size()+3);
  coeffT_3 = 0;
  
  for(int i=0; i< Basis_CE.size() ; i++)
  {
    coeffT_1[i] = phi_1[i];
    coeffT_2[i] = phi_2[i];
    coeffT_3[i] = phi_3[i];
  }      
  
  coeffT_1[Basis_CE.size()] = m_1[0];
  coeffT_1[Basis_CE.size()+1] = m_1[1];
  coeffT_1[Basis_CE.size()+2] = m_1[2];
  coeffT_2[Basis_CE.size()] = m_2[0];
  coeffT_2[Basis_CE.size()+1] = m_2[1];
  coeffT_2[Basis_CE.size()+2] = m_2[2];
  coeffT_3[Basis_CE.size()] = m_3[0];
  coeffT_3[Basis_CE.size()+1] = m_3[1];
  coeffT_3[Basis_CE.size()+2] = m_3[2];
  

  // TEST
//   for(int i = 0; i<Basis_CE.size()+3 ; i++)
//   {
//         if(x_1[i] < 1e-10)
//             x_1[i] = 0.0;
//         if(x_2[i] < 1e-10)
//             x_2[i] = 0.0;
//         if(x_3[i] < 1e-10)
//             x_3[i] = 0.0;
// 
//   }

//   printvector(std::cout, coeffT_1  , "coeffT_1", "--" );
//   printvector(std::cout,  x_1 , "x_1", "--" );  

  
  /////////////////////////////////////////////////////////
  // Create Containers for Basis-Matrices, Correctors and Coefficients 
  /////////////////////////////////////////////////////////
  const std::array<Func2Tensor, 3> x3MatrixBasis = {x3G_1, x3G_2, x3G_3};
  const std::array<VectorCT, 3> coeffContainer = {x_1, x_2, x_3};                                      
//   const std::array<VectorCT, 3> coeffContainer = {coeffT_1, coeffT_2, coeffT_3};                    // TEST
  const std::array<VectorCT, 3> loadContainer = {load_alpha1, load_alpha2, load_alpha3};
//   const std::array<decltype(localPhi_1)*, 3> phiContainer= {&localPhi_1, &localPhi_2, &localPhi_3};    // Achtung hie wurde kein Integralmittelabgezogen!!!!! 
  const std::array<MatrixRT*, 3 > mContainer = {&M_1 , &M_2, & M_3};



  /////////////////////////////////////////////////////////
  // Compute effective quantities
  /////////////////////////////////////////////////////////
  MatrixRT Q(0);

  VectorCT tmp1,tmp2;

  FieldVector<double,3> B_hat ;


  // compute Q 
  for(size_t a = 0; a < 3; a++)
    for(size_t b=0; b < 3; b++)
    {
      assembleCellLoad(Basis_CE, muLocal, lambdaLocal, gamma, tmp1 ,x3MatrixBasis[b]);   // <L i(M_alpha) + sym(grad phi_alpha), i(x3G_beta) >
                                                                                         //entspricht load_alpha1,2,3..
//       printvector(std::cout, tmp1, "tmp1 ", "--"); 
//       printvector(std::cout, load_alpha1, "load_alpha1", "--" );
//       printvector(std::cout, loadContainer[b], "loadContainer[b]", "--" );

      double GGterm = 0.0;
      GGterm = energy(Basis_CE, muLocal, lambdaLocal, x3MatrixBasis[a] , x3MatrixBasis[b]  );   // <L i(x3G_alpha) , i(x3G_beta) >        

      Q[a][b] =  (coeffContainer[a]*tmp1) + GGterm;       // seems symmetric... check positiv definitness?
//       Q[a][b] =  (coeffContainer[a]*loadContainer[b]) + GGterm;       // TEST 

      std::cout << "GGTerm:" << GGterm << std::endl;
      std::cout << "coeff*tmp: " << coeffContainer[a]*tmp1 << std::endl;
//       std::cout << "Test : " << coeffContainer[a]*loadContainer[b] << std::endl;  //same... 
    }
  printmatrix(std::cout, Q, "Matrix Q", "--");


  // compute B_hat
  for(size_t a = 0; a < 3; a++)
  {

    assembleCellLoad(Basis_CE, muLocal, lambdaLocal, gamma, tmp2 ,B_Term); // <L i(M_alpha) + sym(grad phi_alpha), B >

    auto GGterm = energy(Basis_CE, muLocal, lambdaLocal, x3MatrixBasis[a] , B_Term); // <L i(x3G_alpha) , B>

    B_hat[a] = (coeffContainer[a]*tmp2) + GGterm;
    std::cout <<"B_hat[i]: " <<  B_hat[a] << std::endl;
  }




  // compute effective Prestrain
  FieldVector<double, 3> Beff;


  Q.solve(Beff,B_hat);

  std::cout << "Beff : " << std::endl;
  for(size_t i=0; i<3; i++)
  {
    std::cout <<"Beff[i]: " << Beff[i] << std::endl;
  }


  // compute q1,q2,q3
  FieldVector<double ,3> g1 = {1.0 , 0 , 0};
  FieldVector<double ,3> g2 = {0 , 1.0 , 0};
  FieldVector<double ,3> g3 = {0 , 0,  1.0};
  FieldVector<double ,3> tmp_1;
  FieldVector<double ,3> tmp_2;
  FieldVector<double ,3> tmp_3;
  // auto tmp = g1- B_hat;
  // std::cout<< tmp << std::endl;

  Q.mv(g1,tmp_1);
  auto q1_c = tmp_1 * g1;
  std::cout<< "q1_c: " << q1_c << std::endl;

  Q.mv(g2,tmp_2);
  auto q2_c = tmp_2 * g2;
  std::cout<< "q2_c: " << q2_c << std::endl;

  Q.mv(g3,tmp_3);
  auto q3_c = tmp_3 * g3;
  std::cout<< "q3_c: " << q3_c << std::endl;


  std::cout << "Gamma: " << gamma << std::endl;

  log << "q1: " << q1_c << std::endl;
  log << "q2: " << q2_c << std::endl;
  log << "q3: " << q3_c << std::endl;
  log << "b1: " << Beff[1] << std::endl;
  log << "b2: " << Beff[2] << std::endl;
  log << "b3: " << Beff[3] << std::endl;
  log << "b1_hat: " << B_hat[1] << std::endl;
  log << "b2_hat: " << B_hat[2] << std::endl;
  log << "b3_hat: " << B_hat[3] << std::endl;


  
  
  
  //////////////////////////////////////////////////////////////
  // Define Analytic Solutions
  //////////////////////////////////////////////////////////////
  
  // double b1 = (mu1*p1/4)*(beta/(theta+(1-theta)*beta))*(1-theta*(1+alpha));
  // double b2 = (mu1*p1/8)*(1-theta*(1+beta*alpha));

  double b1 = (-(theta/4.0)*mu1*mu2)/(theta*mu1+(1.0-theta)*mu2);
  double b2 = -(theta/4.0)*mu2;
  double b3 = 0.0;
  double q1 = (mu1/6.0)*mu2/(theta*mu1+ (1.0- theta)*mu2);
  double q2 = ((1.0-theta)*mu1+theta*mu2)/6.0;

  std::cout << " --- print analytic solutions --- " << std::endl;
  std::cout << "b1 : " << b1 << std::endl;
  std::cout << "b2 : " << b2 << std::endl;
  std::cout << "b3 : " << b3 << std::endl;
  std::cout << "q1 : " << q1 << std::endl;
  std::cout << "q2 : " << q2 << std::endl;
  std::cout << "q3 should be between q1 and q2"  << std::endl;

  
  
  
  // TODO Define sym grad phi_1  
  // - how to compute <mu>_h ? 
  
//   Func2Tensor symPhi_1_analytic = [] (const Domain& x) {
//                         return MatrixRT{{  (mu1*mu2/((theta*mu1 +(1-theta)*mu2)*muTerm(x)) - 1)*x[2] , 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
//                       };
                      
    auto symPhi_1_analytic = [mu1, mu2, theta, muTerm] (const Domain& x) {
                        return MatrixRT{{  (mu1*mu2/((theta*mu1 +(1-theta)*mu2)*muTerm(x)) - 1)*x[2] , 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
                      };
                      
    auto zeroFunction = [] (const Domain& x) {
                    return MatrixRT{{0.0 , 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};
                    };
                      
                      
    
    FieldVector<double,3> testvector = {1.0/4.0 , 0.0 , 0.25};
//     std::cout << "t[2]: " << testvector[2] << std::endl;
//     std::cout << "muTerm(t):" << muTerm(testvector) << std::endl;
//     auto Teest = symPhi_1_analytic(testvector);
//     printmatrix(std::cout, Teest, "symPhi_1_analytic(t)", "--");
    
    
    
    
  auto L2error = computeL2Error(Basis_CE,phi_1,gamma,symPhi_1_analytic);
  std::cout << "L2-Error: " << L2error << std::endl;
  
  auto L2Norm_Symphi = computeL2Error(Basis_CE,phi_1,gamma,zeroFunction);            // Achtung Norm SymPhi_1
  std::cout << "L2-Norm(Symphi_1): " << L2Norm_Symphi << std::endl;  // TODO Not Needed
  
  VectorCT zeroVec;
  zeroVec.resize(Basis_CE.size());
  zeroVec = 0;

  auto L2Norm_analytic = computeL2Error(Basis_CE,zeroVec ,gamma,symPhi_1_analytic);
  std::cout << "L2-Norm(analytic): " << L2Norm_analytic << std::endl;

  //////////////////////////////////////////////////////////////////////////////////////////////
  // Write result to VTK file
  //////////////////////////////////////////////////////////////////////////////////////////////
  
  /////////////////////////////////////// TEST //////////////////////////////////////
//     auto TestBasis = makeBasis(
//       gridView_CE,
//       power<dim>(
//         lagrange<1>(),
//         flatLexicographic())); 
//     
//     
//   auto& SubPreBasis = Basis_CE.preBasis().subPreBasis();
// 
//   VectorCT fullcoeff(TestBasis.size());
//   
//   std::cout << "TestBasis.size(): "<< TestBasis.size() << std::endl;
//   auto testlocalView = TestBasis.localView();
//   auto perlocalView = Basis_CE.localView();
  
//   std::cout << "testlocalView.size(): " << testlocalView.size() << std::endl;
//   std::cout << "perlocalView.size(): " << perlocalView.size() << std::endl;
//   for (const auto& element : elements(gridView_CE))
//   {
//       testlocalView.bind(element);
//       perlocalView.bind(element);
// //       std::cout << "testlocalView.size(): " << testlocalView.size() << std::endl;
// //       std::cout << "perlocalView.size(): " << perlocalView.size() << std::endl;
//       
//       for(size_t i=0; i<testlocalView.size(); i++)
//       {
//           auto testIdx = testlocalView.index(i);
//           auto perIdx = perlocalView.index(i);
// //             std::cout << "testIdx: " << testIdx << std::endl;
// //             std::cout << "perIdx: " << perIdx << std::endl;
//             fullcoeff[testIdx] = phi_1[perIdx];
//       }
// 
//   }
//   
//   
//   
  
//   for (size_t i = 0; i < TestBasis.size()/3; i++)
//   {
// //       auto c = Basis_CE.indices
// //       auto c = Basis_CE.preBasis_.subPreBasis_.transformation_.mappedIdx_;
// // //       Basis_CE.transformIndex(i);
//       
//         std::cout << "i: "<< i << std::endl;
//         Dune::Functions::FlatMultiIndex<unsigned long> idx = {i};
//         SubPreBasis.transformIndex(idx);
//         std::cout << "idx: " << idx << std::endl;
//         
//         fullcoeff[i] = phi_1[idx];
//         fullcoeff[i] = phi_1[idx];
// 
//   }
//   
  
  

//   printvector(std::cout, fullcoeff, "fullcoeff" , "--");
//   printvector(std::cout, phi_1, "phi_1" , "--");
  
  
  

//     auto correctorFunction_Test = Functions::makeDiscreteGlobalBasisFunction<SolutionRange>(
//     TestBasis,
//     fullcoeff);
    
  //////////////////////////////////////////////////////////////////////////////////////////////
  //     SubsamplingVTKWriter<GridView> vtkWriter(
  //         gridView,
  //         refinementLevels(2));
    
    
  VTKWriter<GridView> vtkWriter(gridView_CE);
  
  
//     vtkWriter.addVertexData(
//     correctorFunction_Test,
//     VTK::FieldInfo("corrector Test", VTK::FieldInfo::Type::vector, dim));
  

  vtkWriter.addVertexData(
    correctorFunction_1,
    VTK::FieldInfo("corrector phi_1", VTK::FieldInfo::Type::vector, dim));
  vtkWriter.addVertexData(
    correctorFunction_2,
    VTK::FieldInfo("corrector phi_2", VTK::FieldInfo::Type::vector, dim));
  vtkWriter.addVertexData(
    correctorFunction_3,
    VTK::FieldInfo("corrector phi_3", VTK::FieldInfo::Type::vector, dim));

  
  vtkWriter.write("CellProblem-result");
  std::cout << "wrote data to file: CellProblem-result" << std::endl;          // better with string for output name..
  
  log.close();

}
